#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "globals.h"
#include "initialize.h"
#include "scattering.h"
#include "dispersion.h"
#include "double_math.h"
#include "file_handling.h"
#include "occupancies.h"
#include "memmory.h"


double ddummy = 0;
int idummy = 0;

void initialize(char working_folder[n_file_chars]){

	double T_hot; double T_cold; 

	//set physical constants, seed random number, set SI prefix
	initialize_constants();
	
	//read in simulation parameters
	initialize_simulation_parameters("simulation_parameters", working_folder, &T_hot, &T_cold); 

	//allocate memmory for global structures
	initialize_arrays();

	//read in electron parameters
	initialize_electron_parameters("electron_parameters", working_folder);

	//read in phonon parameters
	initialize_phonon_parameters("phonon_parameters",working_folder); //read in phonon parameters (conductivity is unused)
	
	 //initial temperature distribution: phonon w/ T_hot in active region, T_cold in inactive region, electron w/ T_cold throughout
	initialize_temperatures(T_hot, T_cold); 

	//Density of states from 3D parabolic
	edos(); 

	//initial electron energy, momentum, and position according to Fermi-Dirac
	initialize_e_distributions(); 

	//Create scattering rate table -- Only optical phonon scattering considered
	initialize_scattering(); 

	//Create generation rate table -- assume isotropic and proportional to density of inital times final states
	initialize_generation();

	return;
}

void initialize_constants() {

	float_code = "%lf";
	int_code = "%d";

	time_t t;
	srand((unsigned)time(&t));

	ME = ME_1*pow(10.0, ME_2);
	EC = EC_1*pow(10.0, EC_2);
	KB = KB_1*pow(10.0, KB_2);
	HBAR = HBAR_1*pow(10.0, HBAR_2);
	NM = pow(10.0, -9);
	PS = pow(10.0, -12);

	VC = pow(10.0, -26);

	return;
}

void initialize_simulation_parameters(char sim_parameters[n_file_chars], char working_folder[n_file_chars],	double * T_hot, double * T_cold) {


	char file[n_file_chars];

	create_file(file, working_folder, sim_parameters, "", 0);

	read_input_file(&simulation.dx, &idummy, "z_step_nm", float_code, 1, file);
	read_input_file(&ddummy, &(simulation.nx), "number_z", int_code, 1, file);
	read_input_file(&ddummy, &(simulation.n_gen_bins), "number_generation_bins", int_code, 1, file);

	read_input_file(&ddummy, &(simulation.gen_mode), "generation_mode", int_code, 1, file);
	read_input_file(&(simulation.generation_fraction), &idummy, "generation_fraction", float_code, 1, file);

	read_input_file(&simulation.dt, &idummy, "t_step_ps", float_code, 1, file);
	read_input_file(&ddummy, &(simulation.nt), "number_t", int_code, 1, file);
	read_input_file(&ddummy, &(simulation.out_freq), "output_frequency", int_code, 1, file);

	read_input_file(&(simulation.de), &idummy, "e_step_eV", float_code, 1, file);
	read_input_file(&ddummy, &(simulation.ne), "number_e", int_code, 1, file);

	read_input_file(T_hot, &idummy, "hot_temperature", float_code, 1, file);
	read_input_file(T_cold, &idummy, "cold_temperature", float_code, 1, file);

	read_input_file(&ddummy, &(simulation.np), "max_particles", int_code, 1, file);
	read_input_file(&ddummy, &(simulation.init_np), "initial_number_of_particles", int_code, 1, file);

	read_input_file(&ddummy, &(simulation.n_interactions), "number_of_interactions", int_code, 1, file);

	simulation.n_interactions_x2 = simulation.n_interactions * 2;

	simulation.dx = simulation.dx*NM;
	simulation.de = simulation.de*EC;
	simulation.dt = simulation.dt*PS;

	LENGTH = (simulation.dx)*(simulation.nx);

	return;
}


void initialize_arrays() {

	//allocate 1-D arrays

	electrons.ensemble = malloc(simulation.np * sizeof(Particle));

	int_1d_dynamic(&(electrons.unused_particle), simulation.np);
	double_1d_dynamic(&(electrons.n_e_below_e), simulation.ne);
	double_1d_dynamic(&(electrons.dos), simulation.ne);
	double_1d_dynamic(&(electrons.temperature), simulation.nx);

	double_1d_dynamic(&(phonons.E_p), simulation.n_interactions);
	double_1d_dynamic(&(phonons.Mpo), simulation.n_interactions);
	
	int_1d_dynamic(&(outputs.current), simulation.nt);
	int_1d_dynamic(&(outputs.cumulative_current), simulation.nt);
	int_1d_dynamic(&(outputs.n_generated), simulation.nt);

	//allocate 2-D arrays

	double_2d_dynamic(&(phonons.T_pO), simulation.nx, simulation.n_interactions);
	double_2d_dynamic(&(phonons.fp), simulation.nx, simulation.n_interactions);

	double_2d_dynamic(&(tables.scattering), simulation.n_interactions_x2, simulation.ne);
	double_2d_dynamic(&(tables.generation), simulation.n_interactions, simulation.ne);

	double_2d_dynamic(&(outputs.heat_abs), simulation.nx, simulation.nt);
	double_2d_dynamic(&(outputs.heat_gen), simulation.nx, simulation.nt);
	
	return;
}

void initialize_electron_parameters(char electron_parameters[n_file_chars], char working_folder[n_file_chars]) {
	
	char file[n_file_chars];

	create_file(file, working_folder, electron_parameters, "", 0);
	read_input_file(&(electrons.me), &idummy, "mass", float_code, 1, file);
	read_input_file(&(electrons.bandgap), &idummy, "bandgap", float_code, 1, file);

	electrons.me = (electrons.me)*ME;
	electrons.bandgap = (electrons.bandgap)*EC;

	_2M_invH2 = 2.0*(electrons.me) / HBAR / HBAR;
	H2_inv2M = HBAR*HBAR / 2.0 / (electrons.me);
	H_inv2M = HBAR / 2.0 / (electrons.me);
	bandgap_d2 = (electrons.bandgap) / 2.0;

	return;
}

void initialize_phonon_parameters(char phonon_parameters[n_file_chars], char working_folder[n_file_chars]) {
	
	int i;
	char file[n_file_chars];
	
	create_file(file, working_folder, phonon_parameters, "", 0);
	read_input_file((phonons.Mpo), &idummy, "Mpo", float_code, simulation.n_interactions, file);
	read_input_file((phonons.E_p), &idummy, "Epo", float_code, simulation.n_interactions, file);

	for (i = 0; i<simulation.n_interactions; i++) {
		phonons.Mpo[i] = phonons.Mpo[i] * EC;
		phonons.E_p[i] = phonons.E_p[i] * EC;
	}

	return;
}

void initialize_scattering(){
	int i;
	int EpO_to_int;
	double sum=0.0;

	for (i = 0; i<simulation.n_interactions; i++){
		EpO_to_int=(int)(phonons.E_p[i]/simulation.de+0.5);
		calculate_scattering_intra_pO(tables.scattering,i,phonons.Mpo[i],EpO_to_int,electrons.dos);
	}

	for (i = 0; i<simulation.n_interactions_x2; i++){
		sum += phonons.fp[0][i/2]*tables.scattering[i][simulation.ne - 1];
	}
	GAMMA_0=sum*1.1; //self scattering must be faster than sum of all real scattering mechanisms
	invGAMMA_0=1.0/GAMMA_0;
	printf("Self scattering rate: Gamma_0 %g\n", GAMMA_0);

	return;
}

void initialize_generation(){
	
	int i;
	int EpO_to_int;
	int bandgap_to_int=(int)(electrons.bandgap/simulation.de+0.5);

	for (i = 0; i<simulation.n_interactions; i++){
		EpO_to_int=(int)(phonons.E_p[i]/simulation.de+0.5);
		calculate_scattering_inter_pO(tables.generation,i,phonons.Mpo[i],EpO_to_int,electrons.dos,bandgap_to_int);
	}

	return;
}



void initialize_temperatures(double T_hot, double T_cold){
	int i;
	int i_int;
	double fp_o;

	for (i = 0; i < simulation.nx; i++){//simulation.nx
		for (i_int = 0; i_int<simulation.n_interactions; i_int++){

			if (i < simulation.n_gen_bins && i_int == simulation.gen_mode){
				phonons.T_pO[i][i_int] = T_hot;
			}
			else{
				phonons.T_pO[i][i_int] = T_cold;
			}
			
			phonon_occupancy(&fp_o, phonons.E_p[i_int], phonons.T_pO[i][i_int]);
			phonons.fp[i][i_int] = fp_o;
		
		}
		electrons.temperature[i]=T_cold;
	}
	KB_TC = KB*T_cold;

	for (i = 0; i < simulation.nx; i++){//simulation.nx
		i_int = 0;
		//printf("%g %g %g\n", phonons.E_p[i_int] / EC, phonons.T_pO[i][i_int], phonons.fp[i][i_int]);
	}


	return;
}

void initialize_e_distributions(){
	
	int particle;
	int i_en;
	int i_z;

	double en;
	double fermi_en=bandgap_d2;
	double fe;

	double ** n_total;
	
	double r_p;
	double rand_pos;
	double rand_en;
	double rand_pol;
	double rand_azi;

	double_2d_dynamic(&n_total,simulation.ne,simulation.nx);

	for (particle = 0; particle<simulation.np; particle++){
		electrons.unused_particle[particle]=1;
	}

	electrons.n_e_below_e[0] = 0.0;
	for (i_z = 0; i_z < simulation.nx; i_z++){
		en = 0.0;
		for (i_en = 1; i_en < simulation.ne; i_en++){
			en += simulation.de;
			electron_occupancy(&fe, en + fermi_en, electrons.temperature[simulation.nx - 1]);
			n_total[i_en][i_z] = n_total[i_en - 1][i_z] + fe*electrons.dos[i_en];
			if (i_z == simulation.nx - 1){
				electrons.n_e_below_e[i_en] = n_total[i_en][simulation.nx - 1];
			}
		}
	}

	for (particle = 0; particle<simulation.init_np; particle++){
		electrons.unused_particle[particle] = 0;

		r_p=random_number();
		rand_pos=r_p*LENGTH;
		rand_en=random_number();
		rand_pol=random_number()*PI;
		rand_azi=random_number()*PI_d2;

		i_z = (int)(r_p*(simulation.nx - 1) + 0.5);

		en=0.0;
		for (i_en = 1; i_en<simulation.ne; i_en++){
			en+=simulation.de;
			if (rand_en*n_total[simulation.ne- 1][i_z]<n_total[i_en][i_z]){
				en -= random_number()*simulation.de;
				break;
			}
		}

		energy_to_momentum(en);

		electrons.ensemble[particle].energy=en;
		electrons.ensemble[particle].momentum = energy_to_momentum(en)*cos(rand_pol)*cos(rand_azi);
		electrons.ensemble[particle].position = rand_pos;
	}

	return;
}

