
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "occupancies.h"
#include "dispersion.h"
#include "double_math.h"
#include "monteCarlo.h"
#include "boundaries.h"
#include "carrier_statistics.h"
#include "generation.h"
#include "memmory.h"
#include "output.h"

void monte_carlo(){

	int i_position; int i_time; int i_interaction; int particle;
	
	double * t_remaining;
	double collision_time=0.0;
	
	int n_expected = 0; double n_expected_2 = 0; double * generate;
	int n_gen = 0; double * n2; double gen_constant;
	int * n_x; double * n_rounded;

	double_1d_dynamic(&(generate), simulation.nx);
	double_1d_dynamic(&(n2), simulation.nx);
	int_1d_dynamic(&(n_x), simulation.nx);
	double_1d_dynamic(&(n_rounded), simulation.nx);
	double_1d_dynamic(&(t_remaining), simulation.np);

	/*	This section determines the intrinsic carrier concentration based on the optical phonon temperature in each position bin.
		It uses this to determine the expected number of electrons in each bin and then come up with a reasonable (but fast) generation 
		This is not tied to Mpo or analytically determined.*/
	
	n_expected = simulation.init_np / simulation.nx;
	n_expected_2 = pow(n_expected, 2);

	WEIGHT = n_expected / sqrt(DE_CONST*pow(electrons.temperature[simulation.nx - 1], 3)*exp(-electrons.bandgap / KB / electrons.temperature[simulation.nx - 1]));

	for (i_position = 0; i_position < simulation.nx; i_position++){
		n2[i_position] = n_expected_2*exp(1.0 / electrons.temperature[i_position] - 1.0 / phonons.T_pO[i_position][simulation.gen_mode]);
	}
	gen_constant = 2.0*PI *EC / HBAR*simulation.generation_fraction*phonons.Mpo[simulation.gen_mode] * phonons.Mpo[simulation.gen_mode] / WEIGHT/WEIGHT*simulation.dt*simulation.dx;

	printf("Initial concentration %g\nExpected generation rate: %g\n", n_expected/WEIGHT,-gen_constant*(n_expected_2 - n2[0]));

	outputs.cumulative_current[0]=0;
	i_time = 0;
	while (i_time<simulation.nt){

		//integrate heat absorption
		if (i_time){
			for (i_interaction = 0; i_interaction < simulation.n_interactions; i_interaction++){
				outputs.heat_abs[i_interaction][i_time] = outputs.heat_abs[i_interaction][i_time - 1];
			}
		}
		outputs.current[i_time] = 0;
	
		//loop through ensemble
		for (particle = 0; particle<simulation.np; particle++){

			//updated existing particles
			if (!electrons.unused_particle[particle]){

				//make sure they have allowed energies (energy step x number energy bins)
				if (electrons.ensemble[particle].energy / simulation.de >= simulation.ne){
					printf("Expand allowed energy range\n");
					exit(1);
				}

				//each particle is allowed to travel for a time step before moving on to the next particle
				t_remaining[particle] += simulation.dt;
				while (t_remaining[particle]>0){
		
					if (electrons.unused_particle[particle]){
						t_remaining[particle] = 0.0;
						break;
					}

					//A scattering event occurs every collision time, determined by the self scattering rate
					//Free flight occurs between collision events
					//And a scattering even is chosen randomly (while maintaing correct statistics) during an event
					collision_time=-invGAMMA_0*log(1.0-random_number());
					
					if (collision_time<t_remaining[particle]){
						free_flight(particle, collision_time, i_time); //free-flight
						if (!electrons.unused_particle[particle]){
							scattering_event(particle, i_time); //chose random scattering event
						}
					}
					else{
						free_flight(particle, t_remaining[particle], i_time);
					}
					t_remaining[particle] -= collision_time;
				}
			}
			else{
				t_remaining[particle] = 0.0;
			}
		}

		//collect number of superparticles in each bin
		number_of_e(n_x);

		//replace particle in ohmic contact
		if (n_expected - n_x[simulation.nx - 1]>0){

			outputs.current[i_time] += n_expected - n_x[simulation.nx - 1];
			xlength_ohmic(n_expected - n_x[simulation.nx - 1]);

		}

		for (i_position = 0; i_position < simulation.nx; i_position++){

			generate[i_position] = gen_constant*(n2[i_position] - pow(n_x[i_position], 2));
		
		}

		//determine how many electrons to place or remove in each bin
		n_gen = net_generation(generate, i_time, n_rounded);

		//determine current and integrate
		if (i_time){
			outputs.cumulative_current[i_time] = outputs.cumulative_current[i_time - 1] + outputs.current[i_time - 1];
			outputs.n_generated[i_time] = outputs.n_generated[i_time - 1] + n_gen;
		}
		else{
			outputs.n_generated[i_time] = n_gen;
		}
		
		//produce checkpoint information
		if ((i_time + 1) % simulation.out_freq == 0){ // update to stdout how current and generation line up  (large mismatch indicates issue)
			printf("t %3.3g  J_cumulative %i gen_cumulative %i\n", 
				simulation.dt*(i_time + 1) / PS, 
				outputs.cumulative_current[i_time], 
				outputs.n_generated[i_time]);
		}

		//start collection energy and particle density near end of simulation to help average out fluctuation
		if (i_time > simulation.nt*0.95){ 
			collect_distribution();
			collect_n();
		}
		i_time++;
		
	}	

	return;
}

//classical free flight
void free_flight(int particle,double collision_time,int current_time){

	electrons.ensemble[particle].position += electrons.ensemble[particle].momentum*H_inv2M*collision_time;

	//boundary conditions
	if (electrons.ensemble[particle].position<0){
		x0_boundary(particle); //reflection
	}
	else if (electrons.ensemble[particle].position >= LENGTH){
		xlength_boundary(particle); //ohmic
		outputs.current[current_time]--;//track current
	}

	return;
}

//random 1d momentum
double isotropic_k_to_kx(double k){
	double kx;
	double polar_angle=random_number()*PI;
	double azimuthal_angle=random_number()*PI_d2;

	kx=k*cos(polar_angle)*cos(azimuthal_angle);

	return kx;

}

//if non-polar optical scattering event chose, update energy and momentum
void update_particle_npo(int particle,int interaction,int is_ab){
	double k;
	if(is_ab){
		electrons.ensemble[particle].energy+=phonons.E_p[interaction];
	}
	else{
		electrons.ensemble[particle].energy -= phonons.E_p[interaction];
	}
	k=energy_to_momentum(electrons.ensemble[particle].energy);
	electrons.ensemble[particle].momentum=isotropic_k_to_kx(k);
	
	return;
}

//choose scattering event
void scattering_event(int particle, int t){
	double en=electrons.ensemble[particle].energy;
	int pos=(int)(electrons.ensemble[particle].position/simulation.dx);
	int interaction = simulation.n_interactions_x2;
	int interaction_d2;
	int en_low=(int)(en/simulation.de);
	int en_high=(int)(en/simulation.de+1.0);
	int is_ab;
	double em = 0.0;
	double r=random_number();
	double gamma_sum=0.0;

	while(interaction){ //Move through all real interactions to determine if a real scattering event occurs, if it does, which one?

		interaction--;

		interaction_d2 = interaction / 2;

		if(interaction % 2){//odd==emission
			is_ab = 0;
			em = 1.0;
		}
		else{
			is_ab = 1;
			em = 0.0;
		}

		gamma_sum += (em + phonons.fp[pos][interaction_d2])*interpolation(tables.scattering[interaction][en_low], tables.scattering[interaction][en_high], simulation.de, en - en_low*simulation.de);

		if (r*GAMMA_0<gamma_sum){ //if real scattering event

			update_particle_npo(particle,interaction_d2,is_ab); //change particle energy and momentum (only non-polar optical considered)

			if (t){ //track total heat absorbed by electron system
				if (is_ab){
					outputs.heat_abs[interaction_d2][t] += phonons.E_p[interaction_d2];
				}
				else{
					outputs.heat_abs[interaction_d2][t] -= phonons.E_p[interaction_d2];
				}
			}

			break;

		}
	}
	return;
}




